/*
 * FILE:   display_hdstation.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2001-2002 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.4 $
 * $Date: 2002/08/27 21:48:18 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"

#ifdef HAVE_HDSTATION		/* From config.h */

#include "debug.h"
#include "display.h"
#include "display_hdstation.h"
#include "tv.h"
#include "dvs_clib.h"
#include "dvs_fifo.h"

#define HDSP_MAGIC	0x12345678

struct state_hdsp {
	pthread_t		 thread_id;
	sv_handle 		*sv;
	sv_fifo	 		*fifo;
	char			*frame_buffer;
	int			 frame_size;
	sv_fifo_buffer		*fifo_buffer;
	sv_fifo_buffer		*display_buffer;
	sv_fifo_buffer		*tmp_buffer;
	pthread_mutex_t		 lock;
	pthread_cond_t		 boss_cv;
	pthread_cond_t		 worker_cv;
	int			 work_to_do;
	int			 boss_waiting;
	int			 worker_waiting;
	uint32_t		 magic;
	char			*bufs[2];
	int			 bufs_index;
};

static void*
display_thread(void *arg)
{
	struct state_hdsp       *s = (struct state_hdsp *) arg;
	int			 res;

	while (1) {
		pthread_mutex_lock(&s->lock);

		while (s->work_to_do == FALSE) {
			s->worker_waiting = TRUE;
			pthread_cond_wait(&s->worker_cv, &s->lock);
			s->worker_waiting = FALSE;
		}

		s->display_buffer = s->tmp_buffer;
		s->work_to_do     = FALSE;

		if (s->boss_waiting) {
			pthread_cond_signal(&s->boss_cv);
		}
		pthread_mutex_unlock(&s->lock);

		res = sv_fifo_putbuffer(s->sv, s->fifo, s->display_buffer, NULL);
		if (res != SV_OK) {
			debug_msg("Error %s\n", sv_geterrortext(res));
			return NULL;
		}
	}
	return NULL;
}

char *
display_hdstation_getf(void *state)
{
	struct state_hdsp	*s = (struct state_hdsp *) state;
	int			 res;

	assert(s->magic == HDSP_MAGIC);

	/* Prepare the new RTP buffer... */
	res = sv_fifo_getbuffer(s->sv, s->fifo, &s->fifo_buffer, NULL, SV_FIFO_FLAG_VIDEOONLY | SV_FIFO_FLAG_FLUSH);
	if (res != SV_OK) {
		debug_msg("Error %s\n", sv_geterrortext(res));
		return NULL;
	}
	s->bufs_index   = (s->bufs_index + 1) % 2;
	s->frame_buffer = s->bufs[s->bufs_index];
	s->frame_size   = 1280 * 720 * 2;
	s->fifo_buffer->dma.addr = s->frame_buffer;
	s->fifo_buffer->dma.size = s->frame_size;

	return s->frame_buffer;
}

int
display_hdstation_putf(void *state, char *frame)
{
	struct state_hdsp	*s = (struct state_hdsp *) state;

	UNUSED(frame);

	assert(s->magic == HDSP_MAGIC);

	pthread_mutex_lock(&s->lock);
	/* Wait for the worker to finish... */
	while (s->work_to_do) {
		s->boss_waiting = TRUE;
		pthread_cond_wait(&s->boss_cv, &s->lock);
		s->boss_waiting = FALSE;
	}

	/* ...and give it more to do... */
	s->tmp_buffer  = s->fifo_buffer;
	s->fifo_buffer = NULL;
	s->work_to_do  = TRUE;

	/* ...and signal the worker */
	if (s->worker_waiting) {
		pthread_cond_signal(&s->worker_cv);
	}
	pthread_mutex_unlock(&s->lock);

	return TRUE;
}

void *
display_hdstation_init(void)
{
	struct state_hdsp 	*s;
	int 		 	 res;

	/* Start the display thread... */
	s = (struct state_hdsp *) malloc(sizeof(struct state_hdsp));
	s->magic        = HDSP_MAGIC;
	s->frame_size   = 0;
	s->frame_buffer = 0;

	s->sv  = sv_open("");
	res = sv_videomode(s->sv, SV_MODE_SMPTE296_60P | SV_MODE_COLOR_YUV422 | SV_MODE_AUDIO_NOAUDIO);
	if (res != SV_OK) {
		goto error;
	}
	res = sv_sync_output(s->sv, SV_SYNCOUT_DEFAULT | SV_SYNCOUT_OUTPUT_GREEN);
	if (res != SV_OK) {
		goto error;
	}

	res = sv_fifo_init(s->sv, &s->fifo, 0, 1, 1, 0, 0);
	if (res != SV_OK) {
		goto error;
	}
	res = sv_fifo_start(s->sv, s->fifo);
	if (res != SV_OK) {
		goto error;
	}

	pthread_mutex_init(&s->lock, NULL);
	pthread_cond_init(&s->boss_cv, NULL);
	pthread_cond_init(&s->worker_cv, NULL);
	s->work_to_do     = FALSE;
	s->boss_waiting   = FALSE;
	s->worker_waiting = FALSE;
	s->display_buffer = NULL;

	s->bufs[0] = malloc(1280 * 720 * 2);
	s->bufs[1] = malloc(1280 * 720 * 2);
	s->bufs_index = 0;
	memset(s->bufs[0], 0, 1280 * 720 * 2);
	memset(s->bufs[1], 0, 1280 * 720 * 2);

	if (pthread_create(&(s->thread_id), NULL, display_thread, (void *) s) != 0) {
		perror("Unable to create display thread\n");
		return NULL;
	}

	return (void *) s;
error:
	debug_msg("Error %s\n", sv_geterrortext(res));
	return NULL;
}

void
display_hdstation_done(void *state)
{
	struct state_hdsp *s = (struct state_hdsp *) state;

	sv_fifo_free(s->sv, s->fifo);
	sv_close(s->sv);
	free(s);
}

display_colour_t
display_hdstation_colour(void *state)
{
	struct state_hdsp *s = (struct state_hdsp *) state;

	assert(s->magic == HDSP_MAGIC);

	return DC_YUV;
}

struct display_type *
display_hdstation_probe(void)
{
	struct display_type	*dt;

	dt = (struct display_type *) malloc(sizeof(struct display_type));
	if (dt != NULL) {
		dt->id	        = DISPLAY_HDSTATION_ID;
		dt->name        = "hdtv";
		dt->description = "DVS HDstation (720p/60)";
		dt->width       = 1280;
		dt->height      = 720;
		dt->max_images  = 1;
	}
	return dt;
}

#endif

