/*
 * FILE:   display_x11.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2001-2002 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.4 $
 * $Date: 2002/08/27 21:48:18 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "display.h"
#include "display_x11.h"
#include "tcl.h"
#include "tk.h"

#define HD_WIDTH 	1280
#define HD_HEIGHT 	720
#define MAGIC_X11	0xcafebabe

struct state_x11 {
	 Tcl_Interp		*interp;
	 Tk_Window		 vw;
	 Display		*display;
	 Window			 window;
	 GC			 gc;
	 int			 vw_depth;
	 Visual			*vw_visual;
	 XImage			*vw_image;
	 XShmSegmentInfo	 vw_shm_segment;
	 int			 xv_port;
	 uint32_t		 magic;			/* For debugging */
};

#ifdef NDEF
static void 
event_handler_x11(ClientData cd, XEvent *ep)
{
	UNUSED(cd);
	UNUSED(ep);
}
#endif

void *
display_x11_init(void)
{
	return NULL;
#ifdef NDEF
	struct state_x11	*s;
	char			*clargs, buffer[10];

	s = (struct state_x11 *) malloc(sizeof(struct state_x11));
	s->magic   = MAGIC_X11;
	s->interp  = Tcl_CreateInterp();
	s->xv_port = -1;
	clargs = Tcl_Merge(argc - 1, argv + 1);
	Tcl_SetVar(s->interp, "argv", clargs, TCL_GLOBAL_ONLY);
	sprintf(buffer, "%d", argc - 1);
	Tcl_SetVar(s->interp, "argc", buffer, TCL_GLOBAL_ONLY);
	Tcl_SetVar(s->interp, "argv0", argv[0], TCL_GLOBAL_ONLY);
	Tcl_SetVar(s->interp, "tcl_interactive", "0", TCL_GLOBAL_ONLY);

	/* There is no easy way of preventing the init functions from */
	/* loading the library files. Ignore error returns and load   */
	/* built in versions.                                         */
	if (Tcl_Init(s->interp) != TCL_OK) {
		fprintf(stderr, "%s\n", Tcl_GetStringResult(s->interp));
		exit(-1);
	}
	if (Tk_Init(s->interp) != TCL_OK) {
		fprintf(stderr, "%s\n", Tcl_GetStringResult(s->interp));
		exit(-1);
	}

	/* Create a bare window to draw into... */
	s->vw = Tk_CreateWindowFromPath(s->interp, Tk_MainWindow(s->interp), ".vw", 0);
	if (s->vw == 0) {
		abort();
	};
	Tk_SetClass(s->vw, "hdrecv_video");
	Tk_CreateEventHandler(s->vw, ExposureMask|StructureNotifyMask, event_handler_x11, 0);
	Tk_GeometryRequest(s->vw, HD_WIDTH, HD_HEIGHT);
	Tk_MapWindow(s->vw);
	s->vw_depth  = Tk_Depth(s->vw);
	s->vw_visual = Tk_Visual(s->vw);

	/* For now, we only support 24-bit TrueColor displays... */
	if (s->vw_depth != 24) {
		printf("Unable to open display: not 24 bit colour\n");
		return NULL;
	}
	if (s->vw_visual->class != TrueColor) {
		printf("Unable to open display: not TrueColor visual\n");
		return NULL;
	}

	/* Do the shared memory magic... */
	s->vw_image = XShmCreateImage(Tk_Display(s->vw), s->vw_visual, s->vw_depth, ZPixmap, NULL, &s->vw_shm_segment, HD_WIDTH, HD_HEIGHT);
	debug_msg("vw_image                   = %p\n", s->vw_image);
	debug_msg("vw_image->width            = %d\n", s->vw_image->width);
	debug_msg("vw_image->height           = %d\n", s->vw_image->height);
	if (s->vw_image->width != HD_WIDTH) {
		printf("Display does not support %d pixel wide shared memory images\n", HD_WIDTH);
		abort();
	}
	if (s->vw_image->width != HD_HEIGHT) {
		printf("Display does not support %d pixel tall shared memory images\n", HD_WIDTH);
		abort();
	}

	s->vw_shm_segment.shmid    = shmget(IPC_PRIVATE, s->vw_image->bytes_per_line * s->vw_image->height, IPC_CREAT|0777);
	s->vw_shm_segment.shmaddr  = shmat(s->vw_shm_segment.shmid, 0, 0);
	s->vw_shm_segment.readOnly = False;
	debug_msg("vw_shm_segment.shmid       = %d\n", s->vw_shm_segment.shmid);
	debug_msg("vw_shm_segment.shmaddr     = %d\n", s->vw_shm_segment.shmaddr);

	s->vw_image->data = s->vw_shm_segment.shmaddr;

	if (XShmAttach(Tk_Display(s->vw), &s->vw_shm_segment) == 0) {
		printf("Cannot attach shared memory segment\n");
		abort();
	}

	/* Get our window onto the screen... */
	Tcl_Eval(s->interp, "pack .vw");

	s->display = Tk_Display(s->vw);
	s->window  = Tk_WindowId(s->vw);
	s->gc      = Tk_GetGC(s->vw, 0, 0);

	while (Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
		/* Process Tcl/Tk events */
	}

	return (void *) s;
#endif
}

void
display_x11_done(void *state)
{
	struct state_x11 *s = (struct state_x11 *) state;

	assert(s->magic == MAGIC_X11);

	XShmDetach(Tk_Display(s->vw), &(s->vw_shm_segment));
	XDestroyImage(s->vw_image);
	shmdt(s->vw_shm_segment.shmaddr);
	shmctl(s->vw_shm_segment.shmid, IPC_RMID, 0);
}

char *
display_x11_getf(void *state)
{
	struct state_x11 *s = (struct state_x11 *) state;
	assert(s->magic == MAGIC_X11);
	return s->vw_image->data;
}

int
display_x11_putf(void *state, char *frame)
{
	struct state_x11 *s = (struct state_x11 *) state;

	assert(s->magic == MAGIC_X11);
	assert(frame == s->vw_image->data);
	XShmPutImage(s->display, s->window, s->gc, s->vw_image, 0, 0, 0, 0, s->vw_image->width, s->vw_image->height, True);
	XFlush(s->display);
	while (Tcl_DoOneEvent(TCL_DONT_WAIT | TCL_ALL_EVENTS)) {
		/* Process Tcl/Tk events */
	}
	return 0;
}

display_colour_t
display_x11_colour(void *state)
{
	struct state_x11 *s = (struct state_x11 *) state;
	assert(s->magic == MAGIC_X11);
	return DC_YUV;
}

struct display_type *
display_x11_probe(void)
{
	struct display_type	*dt;

	dt = (struct display_type *) malloc(sizeof(struct display_type));
	if (dt != NULL) {
		dt->id	        = DISPLAY_X11_ID;
		dt->name        = "x11";
		dt->description = "X Window System";
		dt->width       = 0;	/* FIXME */
		dt->height      = 0;	/* FIXME */
		dt->max_images  = 0;
	}
	return dt;
}

