/*
 * FILE:   main.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2001-2002 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.21 $
 * $Date: 2002/11/28 20:11:34 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "ui.h"
#include "display.h"
#include "video_types.h"
#include "vidcap.h"
#include "rtp.h"
#include "rtp_callback.h"
#include "tv.h"
#include "transmit.h"
#include "tfrc.h"
#include "version.h"

#define EXIT_FAIL_USAGE		1
#define EXIT_FAIL_UI   		2
#define EXIT_FAIL_DISPLAY	3
#define EXIT_FAIL_CAPTURE	4
#define EXIT_FAIL_NETWORK	5
#define EXIT_FAIL_TRANSMIT	6

static int	should_exit  = FALSE;
uint32_t  	RTT=0;    /* this is computed by handle_rr in rtp_callback */
char 		*frame          = NULL;
struct display	*display_device = NULL;

#ifndef WIN32
static void
signal_handler(int signal)
{
	debug_msg("Caught signal %d\n", signal);
	if (signal == SIGINT) {
		should_exit = TRUE;
		return;
	}
	abort();
}
#endif

static void 
usage(void) 
{
	printf("Usage: uv [-d <display_device>] [-t <capture_device>] [-m <mtu>] address\n");
}

static struct display *
initialize_video_display(char *requested_display)
{
	struct display		*d;
	struct display_type	*dt;
	display_id_t		 id = display_get_null_device_id();
	int			 i;

	display_init_devices();
	for (i = 0; i < display_get_device_count(); i++) {
		dt = display_get_device_details(i);
		if (strcmp(requested_display, dt->name) == 0) {
			id = dt->id;
		}
	}
	display_free_devices();

	d = display_init(id);
	if (d != NULL) {
		frame = display_get_frame(d);
	}
	return d;
}

static struct vidcap *
initialize_video_capture(char *requested_capture)
{
	struct vidcap_type	*vt;
	vidcap_id_t		 id = vidcap_get_null_device_id();
	int			 i;

	vidcap_init_devices();
	for (i = 0; i < vidcap_get_device_count(); i++) {
		vt = vidcap_get_device_details(i);
		if (strcmp(vt->name, requested_capture) == 0) {
			id = vt->id;
		}
	}
	vidcap_free_devices();

	return vidcap_init(id, 60);
}

static struct rtp *
initialize_network(char *addr, struct tfrc *tfrc_state)
{
	struct rtp 	*r;
	double		 rtcp_bw = 1024;	/* FIXME */
	
	r = rtp_init(addr, 5004, 5004, 255, rtcp_bw, rtp_recv_callback, (void *) tfrc_state);
	if (r != NULL) {
		rtp_set_option(r, RTP_OPT_WEAK_VALIDATION,   TRUE);
		rtp_set_option(r, RTP_OPT_REUSE_PACKET_BUFS, TRUE);
		rtp_set_option(r, RTP_OPT_PEEK,              64);
		rtp_set_sdes(r, rtp_my_ssrc(r), RTCP_SDES_TOOL, UV_VERSION, strlen(UV_VERSION));
	}
	return r;
}

static struct video_tx *
initialize_transmit(unsigned requested_mtu)
{
	/* Currently this is trivial. It'll get more complex once we */
	/* have multiple codecs and/or error correction.             */
	return tx_init(requested_mtu);
}

int 
main(int argc, char *argv[])
{
	struct rtp		*network_device;
	struct vidcap		*capture_device;
	struct timeval		 timeout, start_time, curr_time, prev_time, prev_rtcp_time;
	uint32_t		 ts = 0;
	int			 ch;
	double			 fps = 60;	/* FIXME */
	struct video_frame	*tx_frame;
	struct video_tx		*tx;
	struct tfrc		*tfrc_state;
	char			*requested_display = "none";
	char			*requested_capture = "none";
	unsigned 		 requested_mtu     = 1500;
	struct sched_param	 sp;

	while ((ch = getopt(argc, argv, "d:t:m:v")) != -1) {
		switch (ch) {
		case 'd' :
			requested_display = optarg;
			break;
		case 't' :
			requested_capture = optarg;
			break;
		case 'm' :
			requested_mtu = atoi(optarg);
			break;
		case 'v' :
        		printf("%s\n", UV_VERSION);
			return EXIT_SUCCESS;
		default :
			usage();
			return EXIT_FAIL_USAGE;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 1) {
		usage();
		return EXIT_FAIL_USAGE;
	}

	gettimeofday(&start_time, NULL);
	gettimeofday(&prev_time, NULL);
	gettimeofday(&prev_rtcp_time, NULL);

	tfrc_state = tfrc_init(start_time);

	if (ui_init(argc, argv) != 0) {
		return EXIT_FAIL_UI;
	}

	display_device = initialize_video_display(requested_display);
	if (display_device == NULL) {
		printf("Unable to open display device: %s\n", requested_display);
		return EXIT_FAIL_DISPLAY;
	}

	capture_device = initialize_video_capture(requested_capture);
	if (capture_device == NULL) {
		printf("Unable to open capture device: %s\n", requested_capture);
		return EXIT_FAIL_CAPTURE;
	}

	network_device = initialize_network(argv[0], tfrc_state);
	if (network_device == NULL) {
		printf("Unable to open network\n");
		return EXIT_FAIL_NETWORK;
	}

	tx = initialize_transmit(requested_mtu);
	if (tx == NULL) {
		printf("Unable to initialize transmitter\n");
		return EXIT_FAIL_TRANSMIT;
	}

#ifndef WIN32
	signal(SIGINT, signal_handler);

	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	if (sched_setscheduler(0, SCHED_FIFO, &sp) != 0) {
		debug_msg("Unable to enable realtime scheduling\n");
	}
#endif

	while (!should_exit) {
		/* Housekeeping and RTCP... */
		gettimeofday(&curr_time, NULL);
		ts = tv_diff(curr_time, start_time) * 90000;
		if (tv_diff(curr_time, prev_rtcp_time) > 1.0) {
			gettimeofday(&prev_rtcp_time, NULL);
			rtp_update(network_device);
			rtp_send_ctrl(network_device, ts, 0);
		}

		/* Receive packets from the network... The timeout is adjusted */
		/* to match the video capture rate, so the transmitter works.  */
		timeout.tv_sec  = 0;
		timeout.tv_usec = ((1.0/fps) - tv_diff(curr_time, prev_time)) * 1000000;
		if (timeout.tv_usec < 10000) {
			timeout.tv_usec = 0;
		}
		rtp_recv(network_device, &timeout, ts);
		gettimeofday(&prev_time, NULL);

		if (tfrc_feedback_is_due(tfrc_state, curr_time)) {
			debug_msg("tfrc rate %f\n", tfrc_feedback_txrate(tfrc_state, curr_time));
		}

		/* Decode and render video... */

		/* Capture and transmit video... */
		tx_frame = vidcap_grab(capture_device);
		if (tx_frame != NULL) {
			tx_send(tx, tx_frame, network_device);
			//free(tx_frame->data);
			free(tx_frame);
		}

	}

	tfrc_done(tfrc_state);
	tx_done(tx);
	rtp_done(network_device);
	vidcap_done(capture_device);
	display_done(display_device);
	ui_done();
	return EXIT_SUCCESS;
}

