/*
 * FILE:   rtp_callback.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2001-2002 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2002/11/06 00:04:28 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "colour_convert.h"
#include "display.h"
#include "ntp.h"
#include "rtp.h"
#include "rtp_callback.h"
#include "tfrc.h"

#define HD_WIDTH 1280
#define HD_HEIGHT 720

extern char		*frame;
extern struct display	*display_device;

char		 	 hdr_buf[100];
struct msghdr		 msg;
struct iovec		 iov[10];

typedef struct {
	uint16_t        y_offset;       /* pixels */
	uint16_t        x_offset;       /* pixels */
	uint16_t        length;         /* pixels */
	uint16_t        flags;          
} payload_hdr_t;

static void
peek_rtp_data(struct rtp *session, rtp_event *e)
{
	rtp_packet		*pckt   = (rtp_packet *) e->data;
	char			*offset = (char *) pckt->data;
	payload_hdr_t		*curr_hdr;
	payload_hdr_t		*hdr[10];
	int			 hdr_count = 0, i;
	int		 	 frame_offset = 0;
	static display_colour_t	 dcm;
	static int		 dcm_done = FALSE;

	curr_hdr = (payload_hdr_t *) pckt->data;
	while (1) {
		offset += 8;
		hdr[hdr_count++] = curr_hdr;
		if ((ntohs(curr_hdr->flags) & (1<<15)) != 0) {
			/* Last header... */
			break;
		}
		if (hdr_count == 10) {
			/* Out of space... */
			break;
		}
		curr_hdr++;
	}

	if (dcm_done == FALSE) {
		dcm = display_get_colour_mode(display_device);
		dcm_done  = TRUE;
	}

	switch (dcm) {
	case DC_YUV:
		/* Display is native YUV. Setup the iovec for recvmsg() */
		msg.msg_name       = NULL;
		msg.msg_namelen    = 0;
		msg.msg_iov        = iov;
		msg.msg_iovlen     = hdr_count + 1;
		msg.msg_control    = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags      = 0;

		/* The RTP and payload headers go first... */
		msg.msg_iov[0].iov_base = hdr_buf;
		msg.msg_iov[0].iov_len  = 12 + (hdr_count * 8);	/* FIXME - assumes no CSRC, etc */
		/* ...then the video data, directly into the frame buffer */
		for (i = 1; i < hdr_count + 1; i++) {
			frame_offset = ((ntohs(hdr[i - 1]->x_offset) + (ntohs(hdr[i - 1]->y_offset) * HD_WIDTH))) * 2;
			msg.msg_iov[i].iov_base = frame + frame_offset;
			msg.msg_iov[i].iov_len  = ntohs(hdr[i - 1]->length);
		}
		rtp_set_recv_iov(session, &msg);
		break;
	case DC_RGB:
		/* Display is RGB, convert the colours and copy over the frame... */
		abort();	/* Not yet implemented */
		break;
	case DC_NULL:
		break;
	}
}

static void
process_rtp_data_iov(struct rtp *session, rtp_event *e)
{
	rtp_packet		*pckt   = (rtp_packet *) e->data;

	UNUSED(session);

	if (pckt->m) {
		display_put_frame(display_device, frame);
		frame = display_get_frame(display_device);
	}
}

extern uint32_t  RTT;

static void
handle_rr(struct rtp *session, rtp_event *e)
{
	float		fract_lost, tmp;
        uint32_t	ntp_sec, ntp_frac, now; 
	rtcp_rr		*r = (rtcp_rr *) e->data;

	if (e->ssrc == rtp_my_ssrc(session)) {
		/* Filter out loopback reports */
		return;
	}

	fract_lost = (r->fract_lost * 100.0) / 256.0;
	printf("Got RR for 0x%08x loss fraction = %f jitter=%d\n", r->ssrc, fract_lost, r->jitter);

        /* Compute RTT=now-lsr-dlsr */
	if (r->lsr != 0) {
		ntp64_time(&ntp_sec, &ntp_frac);
		now = ntp64_to_ntp32(ntp_sec, ntp_frac); 

		debug_msg("%u %u %u = %u\n", now, r->lsr, r->dlsr, now - r->lsr - r->dlsr);

		if ((now < r->lsr) || (now - r->lsr < r->dlsr)) {
			debug_msg("Bogus RTT (packet arrived before it was sent\n");
		} else {
			tmp = ((float) (now - r->lsr - r->dlsr))/65536.0;  /* RTT in  seconds */
			RTT = tmp*1000000; /* RTT in usec */
		}
	}
	UNUSED(session);
}

void
rtp_recv_callback(struct rtp *session, rtp_event *e)
{
	rtcp_app	*pckt_app   = (rtcp_app *)   e->data;
	//rtp_packet	*pckt_rtp   = (rtp_packet *) e->data;
	struct tfrc	*tfrc_state = (struct tfrc *) rtp_get_userdata(session);
	struct timeval	 curr_time;

	gettimeofday(&curr_time, NULL);

	switch (e->type) {
	case PEEK_RTP:
		peek_rtp_data(session, e);
		break;
	case RX_RTP:
		/* FIXME: this is never called, so the TFRC code must move... */
		//tfrc_recv_data(tfrc_state, curr_time, pckt_rtp->seq, pckt_rtp->data_len + 40);
		break;
	case RX_RTP_IOV:
		process_rtp_data_iov(session, e);
		break;
	case RX_RTCP_START:
		break;
	case RX_RTCP_FINISH:
		break;
	case RX_SR:
		break;
	case RX_RR:
		handle_rr(session, e);
		break;
	case RX_RR_EMPTY:
		break;
	case RX_SDES:
		break;
	case RX_APP:
		if (strncmp(pckt_app->name, "RTT_", 4) == 0) {
			assert(pckt_app->length  = 3);
			assert(pckt_app->subtype = 0);
			tfrc_recv_rtt(tfrc_state, curr_time, ntohl(*((int *) pckt_app->data)));
		}
		break;
	case RX_BYE:
		break;
	case SOURCE_DELETED:
		break;
	case SOURCE_CREATED:
		break;
	case RR_TIMEOUT:
		break;
	default:
		debug_msg("Unknown RTP event (type=%d)\n", e->type);
	}
}

